package quimufu.colourful_portals.portal_fluid;

import net.minecraft.block.Block;
import net.minecraft.block.BlockState;
import net.minecraft.block.Blocks;
import net.minecraft.fluid.Fluid;
import net.minecraft.fluid.FluidState;
import net.minecraft.item.Item;
import net.minecraft.registry.Registries;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.sound.SoundEvent;
import net.minecraft.sound.SoundEvents;
import net.minecraft.state.StateManager;
import net.minecraft.state.property.EnumProperty;
import net.minecraft.state.property.IntProperty;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import net.minecraft.util.math.Vec3d;
import net.minecraft.util.math.Vec3i;
import net.minecraft.util.math.random.Random;
import net.minecraft.util.shape.VoxelShape;
import net.minecraft.util.shape.VoxelShapes;
import net.minecraft.world.BlockView;
import net.minecraft.world.World;
import net.minecraft.world.WorldAccess;
import net.minecraft.world.WorldView;
import org.jetbrains.annotations.NotNull;
import quimufu.colourful_portals.ColourfulPortalsMod;
import quimufu.colourful_portals.portal.PortalManager;

import java.util.*;

import static quimufu.colourful_portals.ColourfulPortalsMod.LOGGER;

public class PortalFluid extends Fluid {
    public static final EnumProperty<NullableAxis> AXIS = EnumProperty.of("axis", NullableAxis.class);
    public static final IntProperty AMOUNT = IntProperty.of("amount", 0, 16);

    public final FluidState[] fromBlockStateId;
    private final Map<FluidState, VoxelShape> shapeCache = new IdentityHashMap<>();

    public PortalFluid() {
        super();
        fromBlockStateId = this.getStateManager().getStates().toArray(new FluidState[0]);
        setDefaultState(stateManager.getDefaultState().with(AXIS, NullableAxis.NULL).with(AMOUNT, 16));
    }

    public FluidState unpackFluidState(int i) {
        return fromBlockStateId[i];
    }

    @Override
    protected void appendProperties(StateManager.Builder<Fluid, FluidState> builder) {
        builder.add(AXIS, AMOUNT);
        super.appendProperties(builder);
    }

    @Override
    protected boolean hasRandomTicks() {
        return true;
    }

    @Override
    protected void onRandomTick(World world, BlockPos pos, FluidState state, Random random) {
        if (!world.isClient && !world.getFluidTickScheduler().isQueued(pos, this)) {
            world.scheduleFluidTick(pos, this, this.getTickRate(world));
        }
    }

    @Override
    public Item getBucketItem() {
        return ColourfulPortalsMod.PORTAL_FLUID_BUCKET_ITEM;
    }

    @Override
    protected boolean canBeReplacedWith(FluidState state, BlockView world, BlockPos pos, Fluid fluid, Direction direction) {
        return false;
    }

    @Override
    protected Vec3d getVelocity(BlockView world, BlockPos pos, FluidState state) {
        return calcVelocity(world, pos, state);
    }

    private static Vec3d calcVelocity(BlockView world, BlockPos pos, FluidState state) {
        Vec3d ret = Vec3d.ZERO;
        for (Direction direction : Direction.values()) {
            if (direction.getAxis() != state.get(AXIS).getAxis()) {
                BlockState neighbour = world.getBlockState(pos.offset(direction));
                if (neighbour.isOf(ColourfulPortalsMod.PORTAL_FLUID_BLOCk)
                        && neighbour.getFluidState().get(AMOUNT) == 0) {
                    ret = ret.add(Vec3d.of(direction.getVector()));
                }
            }
        }
        return ret.normalize();
    }

    @Override
    protected void onScheduledTick(World world, BlockPos pos, FluidState state) {
        world.getProfiler().push("portalFluidTick");

        FluidState newStateMe = state;

        int amount = state.get(AMOUNT);
        Direction.Axis currentAxis = state.get(AXIS).getAxis();

        //copy target dir from neighbours to allow easy extension of planes
        if (currentAxis == null) {
            for (Direction direction : Direction.values()) {
                BlockPos offset = pos.offset(direction);
                FluidState fluidState = world.getBlockState(offset).getFluidState();
                if (fluidState.isOf(this) && fluidState.get(AXIS) != NullableAxis.NULL) {
                    currentAxis = fluidState.get(AXIS).getAxis();
                    newStateMe = state.with(AXIS, NullableAxis.of(currentAxis));
                    world.scheduleFluidTick(offset, this, getTickRate(world));
                    break;
                }
            }
        }


        Direction[] costToDirArray = calculateCostToDirArray(world, pos, currentAxis);

        Direction targetDir = null;
        int cheapestCost = 0;
        for (int i = 0; i < costToDirArray.length; i++) {
            Direction direction = costToDirArray[i];
            if (direction != null) {
                targetDir = direction;
                cheapestCost = i;
                break;
            }
        }
        if (currentAxis == null && targetDir != null) {
            currentAxis = getAxis(world, pos, costToDirArray, targetDir, cheapestCost);
            newStateMe = state.with(AXIS, NullableAxis.of(currentAxis));
        }

        if (amount <= 14) {
            //*might* change my state in world, returned value is what new state should be
            newStateMe = steal(world, pos, newStateMe, currentAxis);
        }
        //my state might have changed.
        amount = newStateMe.get(AMOUNT);

        if (amount == 0) {
            //we couldn't steal anything to keep ourselves alive. This ?should? not happen.
            world.setBlockState(pos, Blocks.AIR.getDefaultState());
            world.getProfiler().pop();
            return;
        }
        if (amount == 1) {
            //we are alive, but can't spread, we need no updates
            //create a portal if applicable
            world.getProfiler().pop();
            return;
        }

        if (targetDir != null) {
            int partnerAmount = amount / 2;
            FluidState newStatePartner = state.with(AXIS, NullableAxis.of(currentAxis)).with(AMOUNT, partnerAmount);
            newStateMe = newStateMe.with(AMOUNT, amount - partnerAmount);

            BlockPos posPartner = pos.offset(targetDir);

            world.scheduleFluidTick(posPartner, this, getTickRate(world));
            world.setBlockState(posPartner, newStatePartner.getBlockState());
        } else if (!world.isClient) {
            for (Direction direction : Direction.values()) {
                if (ColourfulPortalsMod.PORTAL_BLOCKS.contains(Registries.BLOCK.getId(world.getBlockState(pos.offset(direction)).getBlock()))
                        && PortalManager.tryIgnite((ServerWorld) world, pos)) {
                    return;
                }
            }

        }

        FluidState actual = world.getBlockState(pos).getFluidState();
        BlockState blockState = newStateMe.getBlockState();
        if (actual != newStateMe) {
            world.setBlockState(pos, blockState);
            world.updateNeighborsAlways(pos, blockState.getBlock());
            world.scheduleFluidTick(pos, this, getTickRate(world));
        } else if (actual != state && Math.abs(state.get(AMOUNT) - actual.get(AMOUNT)) > 1) {
            //we changed by more than one, informing Neighbors, so they can steal as well
            world.updateNeighborsAlways(pos, blockState.getBlock());
        }
        world.getProfiler().pop();
    }

    private static Direction.Axis getAxis(World world, BlockPos pos, Direction[] costToDirArray, Direction targetDir, int cheapestCost) {
        for (int i = cheapestCost + 1; i < 12; i++) {
            Direction direction = costToDirArray[i];
            if (direction != null && direction != targetDir.getOpposite()) {
                return findPerpendicularAxis(targetDir, direction);
            }
        }
        Direction candidate = null;
        for (Direction dir : Direction.values()) {
            if (dir == targetDir.getOpposite() || dir == targetDir) {
                continue;
            }
            BlockPos neighborOfInspectedLocation = pos.offset(dir);
            BlockState neighbourOfInspected = world.getBlockState(neighborOfInspectedLocation);

            //axis of portal Block takes precedence
            if (ColourfulPortalsMod.PORTAL_BLOCKS.contains(Registries.BLOCK.getId(neighbourOfInspected.getBlock()))) {
                return findPerpendicularAxis(targetDir, dir);
            }
            //if adjacent to any solid Block, it's a candidate
            if (neighbourOfInspected.isSideSolidFullSquare(world, neighborOfInspectedLocation, dir.getOpposite())) {
                candidate = dir;
            }
            if (candidate == null && !neighbourOfInspected.isReplaceable()) {
                candidate = dir;
            }
            if (candidate == null && !neighbourOfInspected.isReplaceable()) {
                candidate = dir;
            }

        }
        return findPerpendicularAxis(targetDir, candidate);
    }

    private static Direction.Axis findPerpendicularAxis(Direction one, Direction two) {
        Vec3i axisVec = one.getVector().crossProduct(two.getVector());
        for (Direction d : Direction.values()) {
            if (d.getVector().equals(axisVec)) {
                return d.getAxis();
            }
        }
        throw new RuntimeException("invalid axis Vector " + axisVec);
    }

    @NotNull
    private FluidState steal(World world, BlockPos pos, FluidState myState, Direction.Axis currentAxis) {
        Integer myAmount = myState.get(AMOUNT);
        LinkedHashMap<BlockPos, FluidState> victims = getConnectedNearBy(world, pos, myState, currentAxis, 1);

        //only I
        if (victims.size() == 1) {
            //can't steal from no one :(
            return myState;
        }

        int sum = 0;
        int count = 0;
        for (FluidState s : victims.values()) {
            int i = s.get(AMOUNT);
            sum += i;
            count++;
        }
        //everyone should have average AMOUNT,
        int average = sum / count;
        //and reminder of them should have one more
        int reminder = sum % count;

        if (average == 0) {
            //can't steal from poor people
            return myState;
        }

        //I want to steal target AMOUNT for myself
        int target = average - myAmount;
        if (reminder > 0) {
            target++;
            reminder--;
        }
        if (target <= 0) {
            //how did we even get here?!?
            return myState;
        }

        int stolen = 0;
        HashMap<BlockPos, FluidState> stolenFrom = new HashMap<>();

        for (Map.Entry<BlockPos, FluidState> victim : victims.entrySet()) {
            FluidState victimState = victim.getValue();
            Integer victimAmount = victimState.get(AMOUNT);
            if (victimAmount > average) {
                if (reminder > 0 && victimAmount == average + 1) {
                    reminder--;
                } else {
                    //needs to be stolen from!
                    int toSteal = Math.min(victimAmount - average, target - stolen);
                    FluidState newVictimState = victimState.with(AMOUNT, victimAmount - toSteal);
                    stolen += toSteal;
                    stolenFrom.put(victim.getKey(), newVictimState);
                }
            }
            if (stolen == target) {
                //Already done!
                break;
            }
        }
        if (stolen != target) {
            LOGGER.error("Unsuccessful in Stealing! we somehow stole {}/{}", stolen, target);
        }
        myState = myState.with(AMOUNT, myAmount + stolen);
        world.setBlockState(pos, myState.getBlockState(), Block.NOTIFY_LISTENERS);
        for (Map.Entry<BlockPos, FluidState> updated : stolenFrom.entrySet()) {
            world.setBlockState(updated.getKey(), updated.getValue().getBlockState(), Block.NOTIFY_LISTENERS);
        }


        return myState;
    }

    @NotNull
    private LinkedHashMap<BlockPos, FluidState> getConnectedNearBy(WorldAccess world, BlockPos pos, FluidState myState, Direction.Axis currentAxis, int minAmount) {
        LinkedHashMap<BlockPos, FluidState> connectedNearBy = new LinkedHashMap<>(16);
        HashSet<BlockPos> borderCurr = new HashSet<>();
        HashSet<BlockPos> borderNext = new HashSet<>();
        connectedNearBy.put(pos, myState);
        borderCurr.add(pos);
        while (connectedNearBy.size() <= 17 && !borderCurr.isEmpty()) {
            for (BlockPos blockPos : borderCurr) {
                for (Direction direction : Direction.values()) {
                    BlockPos pos1 = blockPos.offset(direction);
                    FluidState fluidState;
                    if (!connectedNearBy.containsKey(pos1)
                            && (fluidState = world.getBlockState(pos1).getFluidState()).isOf(this)
                            && fluidState.get(AXIS).getAxis() == currentAxis
                            && fluidState.get(AMOUNT) >= minAmount) {
                        borderNext.add(pos1);
                        connectedNearBy.put(pos1, fluidState);
                    }
                }
            }
            borderCurr.clear();
            HashSet<BlockPos> tmp = borderCurr;
            borderCurr = borderNext;
            borderNext = tmp;
        }
        return connectedNearBy;
    }

    @NotNull
    private Direction[] calculateCostToDirArray(World world, BlockPos pos, Direction.Axis currentAxis) {
        int maxCost = 48;
        Direction[] directionCosts = new Direction[maxCost + 1];
        for (Direction direction : Direction.values()) {
            //if this is not the first time spreading (=>currentAxis != null), we wanna spread perpendicular to AXIS
            if (direction.getAxis() == currentAxis)
                continue;
            //can't spread into walls or our own fluid
            BlockState blockState = world.getBlockState(pos.offset(direction));
            if (!blockState.isReplaceable() || blockState.getFluidState().isOf(this))
                continue;
            int dirCost = 0;
            int distance = 1;
            while (distance <= 15 && dirCost < maxCost) {
                BlockPos inspectedLocation = pos.offset(direction, distance);
                BlockState inspectedState = world.getBlockState(inspectedLocation);
                //we reached a wall. Keeping cost
                if (!inspectedState.isReplaceable()) {
                    //make sure we don't lose equally cheap options
                    while (dirCost < 16 && directionCosts[dirCost] != null) {
                        dirCost++;
                    }
                    directionCosts[dirCost] = direction;
                    break;
                }
                //each distance costs 3
                dirCost += 3;
                boolean discount = false;
                for (Direction dir : Direction.values()) {
                    if (dir.getOpposite() == direction) {
                        continue;
                    }
                    BlockPos neighborOfInspectedLocation = inspectedLocation.offset(dir);
                    BlockState neighbourOfInspected = world.getBlockState(neighborOfInspectedLocation);

                    //if adjacent to a Portal Block, 2 less
                    if (ColourfulPortalsMod.PORTAL_BLOCKS.contains(Registries.BLOCK.getId(neighbourOfInspected.getBlock()))) {
                        dirCost -= 2;
                        discount = false;
                        break;
                    }
                    //if adjacent to any solid Block, 1 less
                    if (neighbourOfInspected.isSideSolidFullSquare(world, neighborOfInspectedLocation, dir.getOpposite())) {
                        discount = true;
                        //maybe a Portal Block in another direction
                        continue;
                    }
                }
                if (discount) {
                    dirCost--;
                }
                if (distance == 15) {
                    //we did not find a block within 16 blocks. The fluid can't travel further
                    // punishing with +3 cost, unless we are already at max cost
                    dirCost = Math.max(dirCost + 3, maxCost);
                    directionCosts[dirCost] = direction;
                }
                distance++;
            }
        }
        return directionCosts;
    }

    @Override
    public int getTickRate(WorldView world) {
        return 50;
    }

    @Override
    protected float getBlastResistance() {
        return 100.0f;
    }

    @Override
    public float getHeight(FluidState state, BlockView world, BlockPos pos) {
        return getHeight(state);
    }

    @Override
    public float getHeight(FluidState state) {
        return state.get(AMOUNT) / 16f + 0.1f;//TODO:remove add
    }

    @Override
    public Optional<SoundEvent> getBucketFillSound() {
        return Optional.of(SoundEvents.ITEM_BUCKET_FILL);
    }

    @Override
    protected BlockState toBlockState(FluidState state) {
        return ColourfulPortalsMod.PORTAL_FLUID_BLOCk.getDefaultState().with(PortalFluidBlock.FLUID_STATE_ID, stateManager.getStates().indexOf(state));
    }

    @Override
    public boolean isStill(FluidState state) {
        return true;
    }

    @Override
    public int getLevel(FluidState state) {
        return state.get(AMOUNT) / 2;
    }

    @Override
    public VoxelShape getShape(FluidState state, BlockView world, BlockPos pos) {
        return this.shapeCache.computeIfAbsent(state, this::calcShape);
    }

    private VoxelShape calcShape(FluidState s) {
        float height = getHeight(s);
        if (height == 0) {
            return VoxelShapes.empty();
        }
        double minX = 0.0;
        double minY = 0.0;
        double minZ = 0.0;
        double maxX = 1.0;
        double maxY = 1.0;
        double maxZ = 1.0;
        switch (s.get(AXIS)) {
            case X -> {
                minX = 0.5 - height / 2F;
                maxX = 0.5 + height / 2F;
            }
            case Y -> {
                minY = 0.5 - height / 2F;
                maxY = 0.5 + height / 2F;
            }
            case Z -> {
                minZ = 0.5 - height / 2F;
                maxZ = 0.5 + height / 2F;
            }
        }
        return VoxelShapes.cuboid(minX, minY, minZ, maxX, maxY, maxZ);
    }

    public boolean tryFillWithFluid(WorldAccess world, BlockPos pos, BlockState state, FluidState fluidState) {

        int toDistribute = fluidState.get(AMOUNT);
        FluidState myState = state.getFluidState();
        LinkedHashMap<BlockPos, FluidState> connectedNearBy = getConnectedNearBy(world, pos, myState, myState.get(AXIS).getAxis(), 1);
        int sum = connectedNearBy.values().stream().mapToInt(f -> 16 - f.get(AMOUNT)).sum();
        if (sum < toDistribute) {
            return false;
        }

        HashMap<BlockPos, FluidState> updated = new HashMap<>();
        for (Map.Entry<BlockPos, FluidState> stateEntry : connectedNearBy.entrySet()) {
            FluidState value = stateEntry.getValue();
            Integer targetAmount = value.get(AMOUNT);
            if (targetAmount < 16) {
                //needs to be filled up!
                int fillAmount = Math.min(16 - targetAmount, toDistribute);

                FluidState newTargetState = value.with(AMOUNT, targetAmount + fillAmount);
                toDistribute -= fillAmount;
                updated.put(stateEntry.getKey(), newTargetState);
            }
            if (toDistribute == 0) {
                //Already done!
                break;
            }
        }
        if (toDistribute != 0) {
            LOGGER.error("Unsuccessful in Filling! we somehow have {} left", toDistribute);
            return false;
        }
        if (!world.isClient()) {
            for (Map.Entry<BlockPos, FluidState> update : updated.entrySet()) {
                world.setBlockState(update.getKey(), update.getValue().getBlockState(), Block.NOTIFY_LISTENERS);
                if (!world.getFluidTickScheduler().isQueued(update.getKey(), this)) {
                    world.scheduleFluidTick(update.getKey(), this, this.getTickRate(world));
                }
                world.updateNeighbors(update.getKey(), update.getValue().getBlockState().getBlock());

            }
        }
        return true;
    }

    public boolean tryDrainFluid(WorldAccess world, BlockPos pos, BlockState state) {
        FluidState myState = state.getFluidState();
        int toGeather = this.getDefaultState().get(AMOUNT);
        LinkedHashMap<BlockPos, FluidState> connectedNearBy = getConnectedNearBy(world, pos, myState, myState.get(AXIS).getAxis(), 1);
        int sum = connectedNearBy.values().stream().mapToInt(f -> f.get(AMOUNT)).sum();
        if (sum < toGeather) {
            return false;
        }

        HashMap<BlockPos, FluidState> updated = new HashMap<>();
        for (Map.Entry<BlockPos, FluidState> stateEntry : connectedNearBy.entrySet()) {
            FluidState value = stateEntry.getValue();
            Integer targetAmount = value.get(AMOUNT);
            if (targetAmount > 0) {
                //needs to be filled up!
                int gatherAmount = Math.min(targetAmount, toGeather);

                FluidState newTargetState = value.with(AMOUNT, targetAmount - gatherAmount);
                toGeather -= gatherAmount;
                updated.put(stateEntry.getKey(), newTargetState);
            }
            if (toGeather == 0) {
                //Already done!
                break;
            }
        }
        if (toGeather != 0) {
            LOGGER.error("Unsuccessful in Draining! we somehow got {} to few", toGeather);
            return false;
        }

        if (!world.isClient()) {
            for (Map.Entry<BlockPos, FluidState> update : updated.entrySet()) {
                if (update.getValue().get(AMOUNT) != 0) {
                    world.setBlockState(update.getKey(), update.getValue().getBlockState(), Block.NOTIFY_LISTENERS);
                    if (!world.getFluidTickScheduler().isQueued(update.getKey(), this)) {
                        world.scheduleFluidTick(update.getKey(), this, this.getTickRate(world));
                    }
                } else {
                    world.setBlockState(update.getKey(), Blocks.AIR.getDefaultState(), Block.NOTIFY_ALL);
                }
            }
        }
        return true;
    }
}
